from tqdm import tqdm
import numpy as np
import torch
import collections
import random
import matplotlib.pyplot as plt

class ReplayBuffer:
    def __init__(self, capacity):
        self.buffer = collections.deque(maxlen=capacity)
        
    def add(self, state, action, reward, next_state, done):
        self.buffer.append((state, action, reward, next_state, done))
        
    def sample(self, batch_size):
        transitions = random.sample(self.buffer, batch_size)
        state, action, reward, next_state, done = zip(*transitions)
        
        return np.array(state), action, reward, np.array(next_state), done
    
    def size(self):
        return len(self.buffer)


# 这个函数写的过于麻烦
def moving_average(a, window_size):
    cumulative_sum = np.cumsum(np.insert(a, 0, 0))
    middle = (cumulative_sum[window_size:] - cumulative_sum[:-window_size]) / window_size
    r = np.arange(1, window_size-1, 2)
    begin = np.cumsum(a[:window_size-1])[::2] / r
    end = (np.cumsum(a[:-window_size:-1])[::2] / r)[::-1]
    return np.concatenate((begin, middle, end))

def train_on_policy_agent(env, agent, num_episodes):
    return_list = []  # 保存每一个回合的回报
    for i in range(10):
        with tqdm(total=int(num_episodes/10), desc=f'Iteration {i}') as pbar:
            for i_episode in range(int(num_episodes/10)):
                episode_return = 0  # 记录本回合的回报
                transition_dict = {'states':[], 'actions':[], 'next_states':[], 'rewards':[], 'dones':[]}

                state, _ = env.reset()
                done = False; step = 1

                while not done:
                    action = agent.take_action(state)
                    next_state, reward, done, truc, _ = env.step(action)
                    step += 1
                    if truc is True: done = True

                    transition_dict['states'].append(state)
                    transition_dict['actions'].append(action)
                    transition_dict['next_states'].append(next_state)
                    transition_dict['rewards'].append(reward)
                    transition_dict['dones'].append(done)

                    state = next_state
                    episode_return += reward

                return_list.append(episode_return)
                agent.update(transition_dict)  # agent更新自己的策略；注意这意味着，没一个回合结束都会更新模型

                if (i_episode+1) % 10 == 0:
                    pbar.set_postfix({'episode': f'{num_episodes/10*i + i_episode+1}', 'return': f'{np.mean(return_list[-10:]):.3f}'})
                pbar.update(1)


def train_off_policy_agent(env, agent, num_episodes, replay_buffer: ReplayBuffer, minimal_size, batch_size):
    return_list = []
    for i in range(10):
        with tqdm(total=int(num_episodes/10), desc=f'Iteration {i}') as pbar:
            for i_episode in range(int(num_episodes/10)):
                episode_return = 0

                state, _ = env.reset()
                done = False
                while not done:
                    action = agent.take_action(state)
                    next_state, reward, done, truc, _ = env.step(action) 
                    if truc is True: done = True

                    replay_buffer.add(state, action, reward, next_state, done)

                    state = next_state
                    episode_return += reward

                    if replay_buffer.size() > minimal_size:   # replay_buffer达到一定量之后，每一个action都进行训练
                        b_s, b_a, b_r, b_ns, b_d = replay_buffer.sample(batch_size)
                        transition_dict = {'states':b_s, 'actions':b_a, 'next_states':b_ns, 'rewards':b_r, 'dones':b_d}
                        agent.update(transition_dict)

                return_list.append(episode_return)
                if (i_episode+1) % 10 == 0:
                    pbar.set_postfix({'episode': f'{num_episodes/10*i+i_episode+1}', 'return': f'{np.mean(return_list[-10:]):.3f}'})
                pbar.update(1)

    return return_list


def compute_advantage(gamma, lmbda, td_delta):
    td_delta = td_delta.detach().numpy()
    advantage_list = []
    advantage = 0.0
    for delta in td_delta[::-1]:
        advantage = gamma * lmbda * advantage + delta
        advantage_list.append(advantage)
    advantage_list.reverse()
    return torch.tensor(advantage_list, dtype=torch.float)


def plot_rewards(return_list, env_name,):
    episodes_list = list(range(len(return_list)))
    plt.plot(episodes_list, return_list)
    plt.xlabel('Episodes')
    plt.ylabel('Returns')
    plt.title('DQN on {}'.format(env_name))
    plt.show()

    mv_return = moving_average(return_list, 9)
    plt.plot(episodes_list, mv_return)
    plt.xlabel('Episodes')
    plt.ylabel('Returns')
    plt.title('DQN on {}'.format(env_name))
    plt.show()